"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const aws_sdk_1 = require("aws-sdk");
const fs_extra_plus_1 = require("fs-extra-plus");
const stream_1 = require("stream");
const bucket_base_1 = require("./bucket-base");
const micromatch = require("micromatch");
async function getAwsBucket(cfg) {
    const credentials = new aws_sdk_1.Credentials(cfg.access_key_id, cfg.access_key_secret);
    // Create S3 service object
    const s3 = new aws_sdk_1.S3({ apiVersion: '2006-03-01', credentials });
    return new AwsBucket(s3, cfg.bucketName);
}
exports.getAwsBucket = getAwsBucket;
class AwsBucket {
    constructor(s3, bucketName) {
        this.s3 = s3;
        this.baseParams = { Bucket: bucketName };
    }
    get type() {
        return 's3';
    }
    get name() {
        return this.baseParams.Bucket;
    }
    getPath(obj) {
        return obj.Key; // TODO: need to investigate when Key is empty in S3. 
    }
    async exists(path) {
        const file = await this.getFile(path);
        return (file) ? true : false;
    }
    async getFile(path) {
        try {
            const object = await this.s3.headObject(Object.assign(Object.assign({}, this.baseParams), { Key: path })).promise();
            const updated = (object.LastModified) ? object.LastModified.toISOString() : undefined;
            return {
                bucket: this,
                path,
                updated,
                size: object.ContentLength,
                contentType: object.ContentType
            };
        }
        catch (ex) {
            //  if NotFound, return false
            if (ex.code === 'NotFound') {
                return null;
            }
            // otherwise, propagate the exception
            else {
                throw ex;
            }
        }
    }
    /**
     *
     * @param path prefix path or glob (the string before the first '*' will be used as prefix)
     */
    async list(prefixOrGlob) {
        const awsFiles = await this.listAwsFiles(prefixOrGlob);
        return awsFiles.map(gf => this.toFile(gf));
    }
    async copy(pathOrGlob, destDir) {
        const awsFiles = await this.listAwsFiles(pathOrGlob);
        const files = await bucket_base_1.commonBucketCopy(this, awsFiles, pathOrGlob, destDir, async (awsFile, dest) => {
            const destAwsBucket = (dest.bucket instanceof AwsBucket) ? dest.bucket : null;
            if (!destAwsBucket) {
                throw new Error(`destBucket type ${dest.bucket.type} does not match source bucket type ${this.type}. For now, cross bucket type copy not supported.`);
            }
            const sourcePath = awsFile.Key;
            const params = {
                CopySource: `${this.name}/${sourcePath}`,
                Bucket: destAwsBucket.name,
                Key: dest.path
            };
            await this.s3.copyObject(params).promise();
        });
    }
    async download(pathOrGlob, localPath) {
        const awsFiles = await this.listAwsFiles(pathOrGlob);
        const files = await bucket_base_1.commonBucketDownload(this, awsFiles, pathOrGlob, localPath, async (object, localPath) => {
            const remotePath = object.Key;
            const params = Object.assign(Object.assign({}, this.baseParams), { Key: remotePath });
            const remoteReadStream = this.s3.getObject(params).createReadStream();
            const localWriteStream = fs_extra_plus_1.createWriteStream(localPath);
            const writePromise = new Promise((resolve, reject) => {
                localWriteStream.once('close', () => {
                    resolve();
                });
                localWriteStream.once('error', (ex) => {
                    reject(ex);
                });
                remoteReadStream.pipe(localWriteStream);
            });
            await writePromise;
        });
        return files;
    }
    async downloadAsText(path) {
        const params = Object.assign(Object.assign({}, this.baseParams), { Key: path });
        //const remoteReadStream = this.s3.getObject(params).createReadStream();
        const obj = await this.s3.getObject(params).promise();
        const content = obj.Body.toString();
        return content;
    }
    async upload(localFileOrDirOrGlob, destPath) {
        return bucket_base_1.commonBucketUpload(this, localFileOrDirOrGlob, destPath, async (localPath, fullDestPath, contentType) => {
            const localFileData = await fs_extra_plus_1.readFile(localPath);
            const awsResult = await this.s3.putObject(Object.assign(Object.assign({}, this.baseParams), { Key: fullDestPath, Body: localFileData, ContentType: contentType })).promise();
            return { bucket: this, path: fullDestPath, size: localFileData.length };
        });
    }
    async uploadContent(path, content) {
        const ContentType = bucket_base_1.getContentType(path);
        await this.s3.putObject(Object.assign(Object.assign({}, this.baseParams), { Key: path, Body: content, ContentType: ContentType })).promise();
    }
    async createReadStream(path) {
        const params = Object.assign(Object.assign({}, this.baseParams), { Key: path });
        const obj = this.s3.getObject(params);
        if (!obj) {
            throw new Error(`Object not found for ${path}`);
        }
        return obj.createReadStream();
    }
    async createWriteStream(path) {
        var pass = new stream_1.PassThrough();
        const params = Object.assign(Object.assign(Object.assign({}, this.baseParams), { Key: path }), { Body: pass });
        this.s3.upload(params);
        return pass;
    }
    async delete(path) {
        if (!path) {
            throw new Error(`AwsBucket - ERROR - Can't delete null or empty path`);
        }
        try {
            process.stdout.write(`Deleting s3://${this.baseParams.Bucket}/${path}`);
            // NOTE: For aws API, the s3.deleteObject seems to return exactly the same if the object existed or not. 
            //       Therefore, we need to do an additional ping to know if the file exist  or not to return true/false
            const exists = await this.exists(path);
            if (exists) {
                // NOTE: between the first test and this delete, the object might have been deleted, but since s3.deleteObjecct
                //       does not seems to tell if the object exits or not, this is the best can do.
                await this.s3.deleteObject(Object.assign(Object.assign({}, this.baseParams), { Key: path })).promise();
                process.stdout.write(` - DONE\n`);
                return true;
            }
            else {
                process.stdout.write(` - Skipped (object not found)\n`);
                return false;
            }
        }
        catch (ex) {
            process.stdout.write(` - FAILED - ABORT - Cause ${ex}\n`);
            throw ex;
        }
    }
    async deleteAll(files) {
        return await bucket_base_1.commonDeleteAll(this, files);
    }
    //#region    ---------- Private ---------- 
    /**
     * List the googleFiles for this bucket;
     */
    async listAwsFiles(prefixOrGlob) {
        const { prefix, glob } = bucket_base_1.parsePrefixOrGlob(prefixOrGlob);
        // build the list params
        let listParams = undefined;
        if (prefix) {
            listParams = { Prefix: prefix };
        }
        const params = Object.assign(Object.assign({}, this.baseParams), listParams);
        // perform the s3 list request
        try {
            const awsResult = await this.s3.listObjects(params).promise();
            const awsFiles = awsResult.Contents;
            // if glob, filter again the result
            let files = (!glob) ? awsFiles : awsFiles.filter(af => micromatch.isMatch(af.Key, glob));
            return files;
        }
        catch (ex) {
            throw ex;
        }
    }
    toFile(awsFile) {
        if (!awsFile) {
            throw new Error(`No awsFile`);
        }
        const updated = (awsFile.LastModified) ? awsFile.LastModified.toISOString() : undefined;
        // FIXME: Needs to handle when Key or Size is undefined.
        return {
            bucket: this,
            path: awsFile.Key,
            size: awsFile.Size,
            updated: updated
        };
    }
}
