/// <reference types="node" />
import { Readable, Writable } from 'stream';
export interface BucketFile {
    bucket: Bucket;
    path: string;
    size?: number;
    updated?: string;
    contentType?: string;
    local?: string;
}
export declare type BucketType = 's3' | 'gs';
export declare type BucketFileDeleted = BucketFile & {
    deleted: boolean;
};
export interface Bucket<F = any> {
    type: BucketType;
    name: string;
    /** Return the path of a cloud "file object" */
    getPath(obj: F): string;
    exists(path: string): Promise<boolean>;
    /**
     * Get and return a BucketFile for a given path (will do a cloud bucket query).
     * Returns null if not found. Throw exception if any other exception than notfound.
     */
    getFile(path: String): Promise<BucketFile | null>;
    list(prefixOrGlob?: String): Promise<BucketFile[]>;
    /**
     * Will copy one or more file to a destination file (then require single match, i.e. no glob),
     * or to a destination folder.
     * @param prefixOrGlob full file name, or prefix or glob. If multiple match, to need dest must be a dir (end with '/')
     * @param dest can be dir path  (copying multiple file and preserving filename and relative dir structure), or full file name (require single match)
     */
    copy(prefixOrGlob: string, dest: string | BucketFile): Promise<void>;
    /**
     * Download one or more remote bucket file to a local file or a folder structure.
     * @param prefixOrGlob
     * @param localDir
     *  If end with '/' then all files from the prefixOrGlob will be downloaded with their originial filename (and relative folder structure).
     *  Otherwise, if full file name, then, make sure there is onyl one matching bucket source, and copy to this file destination (to rename on download)
     */
    download(prefixOrGlob: string, localPath: string): Promise<BucketFile[]>;
    downloadAsText(path: string): Promise<string>;
    upload(localPath: string, path: string): Promise<BucketFile[]>;
    uploadContent(path: string, content: string): Promise<void>;
    createReadStream(path: string): Promise<Readable>;
    createWriteStream(path: string): Promise<Writable>;
    /**
     * Delete a single file.
     * @param path
     */
    delete(path: string): Promise<boolean>;
    deleteAll(files: BucketFile[]): Promise<BucketFileDeleted[]>;
}
/**
 * Build the full destination path from the local path name and the destPath
 * - If `destPath` ends with `/`, then baseName of `localPath` is concatenated.
 * - Otherwise, `destPath` is the fullDestPath.
 *
 * @throws exception if destPath is not present.
 */
export declare function buildFullDestPath(localPath: string, destPath: string): string;
/**
 * Return a clean prefix and glob when defined in the string. Clean prefix, meaning, glob less one,
 * that can be passed to most cloud storage api.
 *
 * @param prefixOrGlob undefined, null, e.g., 'some-prefix', 'folder/', 'folder/glob-pattern.*'
 * @returns {prefix, glob, baseDir}
 * 					- prefix is the first characters unitl the first glob character ('*')
 * 					- glob is prefixOrGlob value if it is a glob, otherwise undefined.
 * 					- baseDir is the eventual longest directory path without any glob char (ending with '/')
 */
export declare function parsePrefixOrGlob(prefixOrGlob?: string): {
    prefix: string | undefined;
    glob: string | undefined;
    baseDir: string | undefined;
};
declare type ItemDownloadFn<F> = (object: F, localPath: string) => Promise<void>;
export declare function commonBucketDownload<F>(bucket: Bucket, cloudFiles: F[], pathOrGlob: string, localPath: string, downloadr: ItemDownloadFn<F>): Promise<BucketFile[]>;
declare type ItemUploadFn = (localFilePath: string, remoteFilePath: string, contentType?: string) => Promise<BucketFile>;
export declare function commonBucketUpload<F>(bucket: Bucket, localFileOrDirOrGlob: string, remotePath: string, uploadr: ItemUploadFn): Promise<BucketFile[]>;
declare type ItemCopyFn<F> = (Object: F, destDir: BucketFile) => Promise<void>;
export declare function commonBucketCopy<F>(bucket: Bucket, cloudFiles: F[], pathOrGlob: string, dest: string | BucketFile, copier: ItemCopyFn<F>): Promise<BucketFile[]>;
export declare function commonDeleteAll(bucket: Bucket, files: BucketFile[]): Promise<BucketFileDeleted[]>;
export declare function getContentType(path: string): string | undefined;
export {};
