"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const Path = require("path");
const fs_extra_plus_1 = require("fs-extra-plus");
const mime = require("mime-types");
//#region    ---------- Common Bucket Utils ---------- 
/**
 * Build the full destination path from the local path name and the destPath
 * - If `destPath` ends with `/`, then baseName of `localPath` is concatenated.
 * - Otherwise, `destPath` is the fullDestPath.
 *
 * @throws exception if destPath is not present.
 */
function buildFullDestPath(localPath, destPath) {
    // we do not have a dest path, throw error
    if (!destPath) {
        throw new Error('No depthPath');
    }
    let fullDestPath;
    // if it is a folder, we just concatinate the base name
    if (destPath.endsWith('/')) {
        const srcBaseName = Path.basename(localPath);
        fullDestPath = destPath + srcBaseName;
    }
    // if the destPath is not a folder, assume it is the new file name.
    else {
        fullDestPath = destPath;
    }
    return fullDestPath;
}
exports.buildFullDestPath = buildFullDestPath;
/**
 * Return a clean prefix and glob when defined in the string. Clean prefix, meaning, glob less one,
 * that can be passed to most cloud storage api.
 *
 * @param prefixOrGlob undefined, null, e.g., 'some-prefix', 'folder/', 'folder/glob-pattern.*'
 * @returns {prefix, glob, baseDir}
 * 					- prefix is the first characters unitl the first glob character ('*')
 * 					- glob is prefixOrGlob value if it is a glob, otherwise undefined.
 * 					- baseDir is the eventual longest directory path without any glob char (ending with '/')
 */
function parsePrefixOrGlob(prefixOrGlob) {
    let glob;
    let prefix;
    let baseDir;
    if (prefixOrGlob && prefixOrGlob.length > 0) {
        const firstWildIdx = prefixOrGlob.indexOf('*');
        // if it has a '*' then it is a pattern
        if (firstWildIdx > 0) {
            glob = prefixOrGlob;
            prefix = prefixOrGlob.substring(0, firstWildIdx);
        }
        // otherwise, it is just a 
        else {
            prefix = prefixOrGlob;
        }
    }
    if (prefix) {
        const lastSlashIdx = prefix.lastIndexOf('/');
        if (lastSlashIdx > -1) {
            baseDir = prefix.substring(0, lastSlashIdx + 1);
        }
    }
    return { prefix, glob, baseDir };
}
exports.parsePrefixOrGlob = parsePrefixOrGlob;
async function commonBucketDownload(bucket, cloudFiles, pathOrGlob, localPath, downloadr) {
    const isLocalPathDir = localPath.endsWith('/');
    // If not a local directory, make sure we have only one file.
    // TODO: might check if the pathOrGlob is a glob as well to prevent it (in case there is only one match)
    if (!isLocalPathDir && cloudFiles.length > 1) {
        throw new Error(`Cannot copy multiple files ${pathOrGlob} to the same local file ${localPath}. Download to a directory (end with '/') to download multipel file.`);
    }
    const files = [];
    const { baseDir } = parsePrefixOrGlob(pathOrGlob);
    for (let cf of cloudFiles) {
        const remotePath = bucket.getPath(cf);
        const localFilePath = (isLocalPathDir) ? getDestPath(baseDir, remotePath, localPath) : localPath;
        const localPathDir = Path.dirname(localFilePath);
        await fs_extra_plus_1.mkdirp(localPathDir);
        process.stdout.write(`Downloading ${bucket.type}://${bucket.name}/${remotePath} to ${localFilePath}`);
        try {
            await downloadr(cf, localFilePath);
            process.stdout.write(` - DONE\n`);
            const file = { bucket, path: remotePath, size: -1, local: localFilePath };
            files.push(file);
        }
        catch (ex) {
            process.stdout.write(` - FAIL - ABORT - Cause: ${ex}\n`);
            throw ex;
        }
    }
    return files;
}
exports.commonBucketDownload = commonBucketDownload;
async function commonBucketUpload(bucket, localFileOrDirOrGlob, remotePath, uploadr) {
    const bucketFiles = [];
    if (localFileOrDirOrGlob.endsWith('/')) {
        localFileOrDirOrGlob = localFileOrDirOrGlob + '**/*.*';
    }
    const isLocalGlob = localFileOrDirOrGlob.includes('*');
    const { baseDir } = parsePrefixOrGlob(localFileOrDirOrGlob);
    const localFiles = await fs_extra_plus_1.glob(localFileOrDirOrGlob);
    for (const localPath of localFiles) {
        // if we have an localFileExpression (globs), then, we build the fullDestPath relative to the baseDir of the glob (last / before the first *)
        const fullDestPath = (isLocalGlob) ? getDestPath(baseDir, localPath, remotePath) : buildFullDestPath(localPath, remotePath);
        const contentType = getContentType(fullDestPath);
        process.stdout.write(`Uploading file ${localPath} to ${bucket.type}://${bucket.name}/${fullDestPath}`);
        try {
            const bucketFile = await uploadr(localPath, fullDestPath, contentType);
            bucketFiles.push(bucketFile);
            process.stdout.write(` - DONE\n`);
        }
        catch (ex) {
            process.stdout.write(` - FAIL - ABORT - Cause: ${ex}\n`);
            throw ex;
        }
    }
    return bucketFiles;
}
exports.commonBucketUpload = commonBucketUpload;
async function commonBucketCopy(bucket, cloudFiles, pathOrGlob, dest, copier) {
    const destBucket = ((typeof dest === 'string') ? bucket : dest.bucket);
    const destPath = (typeof dest === 'string') ? dest : dest.path;
    const isDestPathDir = destPath.endsWith('/');
    // If not a local directory, make sure we have only one file.
    // TODO: might check if the pathOrGlob is a glob as well to prevent it (in case there is only one match)
    if (!isDestPathDir && cloudFiles.length > 1) {
        throw new Error(`Cannot copy multiple files ${pathOrGlob} to the same bucket file ${destPath}. Download to a directory (end with '/') to download multipel file.`);
    }
    const { baseDir } = parsePrefixOrGlob(pathOrGlob);
    const files = [];
    for (let cf of cloudFiles) {
        const remotePath = bucket.getPath(cf);
        const destFilePath = (isDestPathDir) ? getDestPath(baseDir, remotePath, destPath) : destPath;
        process.stdout.write(`Copying ${bucket.type}://${bucket.name}/${remotePath} to ${bucket.type}://${destBucket.name}/${destFilePath}`);
        try {
            await copier(cf, { bucket: destBucket, path: destFilePath });
            process.stdout.write(` - DONE\n`);
        }
        catch (ex) {
            process.stdout.write(` - FAIL - ABORT - Cause: ${ex}\n`);
            throw ex;
        }
    }
    return files;
}
exports.commonBucketCopy = commonBucketCopy;
async function commonDeleteAll(bucket, files) {
    const filesInfo = [];
    // validate that all files are same bucket
    for (const file of files) {
        // check if same bucket
        if (file.bucket !== bucket) {
            throw new Error(`Cannot delete file from another bucket ${bucket.name} should match file bucket ${file.bucket.name}`);
        }
    }
    for (const file of files) {
        const deleted = await bucket.delete(file.path);
        filesInfo.push(Object.assign(Object.assign({}, file), { deleted }));
    }
    return filesInfo;
}
exports.commonDeleteAll = commonDeleteAll;
//#endregion ---------- /Common Bucket Utils ---------- 
function getDestPath(baseDir, remotePath, destPathDir) {
    const baseName = Path.basename(remotePath);
    const filePath = (baseDir) ? Path.relative(baseDir, remotePath) : baseName;
    const destPath = `${destPathDir}${filePath}`;
    return destPath;
}
function getContentType(path) {
    let ct = mime.lookup(path);
    let contentType = (ct) ? ct : undefined;
    return contentType;
}
exports.getContentType = getContentType;
