"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const storage_1 = require("@google-cloud/storage");
const stream_1 = require("stream");
const bucket_base_1 = require("./bucket-base");
const micromatch = require("micromatch");
async function getGcpBucket(cfg) {
    // TODO: valid cfg
    const googleStorageConf = {
        projectId: cfg.project_id,
        credentials: {
            client_email: cfg.client_email,
            private_key: cfg.private_key
        }
    };
    const storage = new storage_1.Storage(googleStorageConf);
    const googleBucket = storage.bucket(cfg.bucketName);
    return new GcpBucket(googleBucket);
}
exports.getGcpBucket = getGcpBucket;
class GcpBucket {
    constructor(googleBucket) {
        this.googleBucket = googleBucket;
    }
    get type() {
        return 'gs';
    }
    get name() {
        return this.googleBucket.name;
    }
    getPath(obj) {
        return obj.name;
    }
    async exists(path) {
        // Note: note gcp has specific file.exists method
        const result = await this.googleBucket.file(path).exists();
        return result[0];
    }
    async getFile(path) {
        const googleFile = this.googleBucket.file(path);
        try {
            const f = (await googleFile.get())[0];
            return this.toFile(f);
        }
        catch (ex) {
            // not found return null, as per getFile design.
            if (ex.code === 404) {
                return null;
            }
            // otherwise, propagate exception 
            else {
                throw ex;
            }
        }
    }
    /**
     *
     * @param path prefix path or glob (the string before the first '*' will be used as prefix)
     */
    async list(prefixOrGlob) {
        const googleFiles = await this.listGoogleFiles(prefixOrGlob);
        return googleFiles.map(gf => this.toFile(gf));
    }
    async copy(pathOrGlob, destDir) {
        const gfiles = await this.listGoogleFiles(pathOrGlob);
        const files = await bucket_base_1.commonBucketCopy(this, gfiles, pathOrGlob, destDir, async (googleFile, dest) => {
            const destGcpBucket = (dest.bucket instanceof GcpBucket) ? dest.bucket : null;
            if (!destGcpBucket) {
                throw new Error(`destBucket type ${dest.bucket.type} does not match source bucket type ${this.type}. For now, cross bucket type copy not supported.`);
            }
            const destFile = destGcpBucket.googleBucket.file(dest.path);
            await googleFile.copy(destFile);
        });
    }
    async download(pathOrGlob, localPath) {
        const googleFiles = await this.listGoogleFiles(pathOrGlob);
        const files = await bucket_base_1.commonBucketDownload(this, googleFiles, pathOrGlob, localPath, async (gf, localPath) => {
            await gf.download({ destination: localPath });
        });
        return files;
    }
    async downloadAsText(path) {
        const googleFile = this.googleBucket.file(path);
        const buffer = await googleFile.download();
        return buffer.toString();
    }
    async upload(localFileOrDirOrGlob, destPath) {
        return bucket_base_1.commonBucketUpload(this, localFileOrDirOrGlob, destPath, async (localPath, remoteFilePath, contentType) => {
            const googleBucket = this.googleBucket;
            const googleFile = (await googleBucket.upload(localPath, { destination: remoteFilePath, contentType }))[0];
            return this.toFile(googleFile);
        });
    }
    async uploadOld(localPath, destPath) {
        const googleBucket = this.googleBucket;
        const fullDestPath = bucket_base_1.buildFullDestPath(localPath, destPath);
        const contentType = bucket_base_1.getContentType(destPath);
        // TODO: Needs to do 
        process.stdout.write(`Uploading file ${localPath} to gs://${this.name}/${fullDestPath}`);
        try {
            const googleFile = (await googleBucket.upload(localPath, { destination: fullDestPath, contentType }))[0];
            process.stdout.write(' - DONE\n');
            return this.toFile(googleFile);
        }
        catch (ex) {
            process.stdout.write(` - FAIL - ABORT - Cause: ${ex}`);
            throw ex;
        }
    }
    async uploadContent(path, content) {
        const googleFile = this.googleBucket.file(path);
        const uploadReadable = new stream_1.Readable();
        const contentType = bucket_base_1.getContentType(path);
        return new Promise(function (resolve, reject) {
            uploadReadable
                .pipe(googleFile.createWriteStream({ contentType }))
                .on('error', function (err) {
                reject(err);
            })
                .on('finish', function () {
                resolve();
            });
            uploadReadable.push(content);
            uploadReadable.push(null);
        });
    }
    async createReadStream(path) {
        const googleFile = this.googleBucket.file(path);
        return googleFile.createReadStream();
    }
    async createWriteStream(path) {
        const googleFile = this.googleBucket.file(path);
        return googleFile.createWriteStream();
    }
    async delete(path) {
        const googleFile = this.googleBucket.file(path);
        process.stdout.write(`Deleting gs://${this.name}/${path}`);
        if (googleFile) {
            try {
                await googleFile.delete();
                process.stdout.write(` - DONE\n`);
            }
            catch (ex) {
                // if not found, just return false.
                if (ex.code === 404) {
                    process.stdout.write(` - Skipped (object not found)\n`);
                    return false;
                }
                else {
                    process.stdout.write(` - FAILED - ABORT - Cause ${ex}\n`);
                    throw ex;
                }
            }
            // TODO: Probably needs to return true only if deleted. 
            return true;
        }
        else {
            return false;
        }
    }
    async deleteAll(files) {
        return await bucket_base_1.commonDeleteAll(this, files);
    }
    //#region    ---------- Private ---------- 
    toFile(googleFile) {
        if (!googleFile) {
            throw new Error(`No googleFile`);
        }
        const size = (googleFile.metadata.size) ? Number(googleFile.metadata.size) : undefined;
        return {
            path: googleFile.name,
            bucket: this,
            size,
            updated: googleFile.metadata.updated,
            contentType: googleFile.metadata.contentType
        };
    }
    /**
     * List the googleFiles for this bucket;
     */
    async listGoogleFiles(prefixOrGlob) {
        // extract the eventual prefix and glob from param
        const { prefix, glob } = bucket_base_1.parsePrefixOrGlob(prefixOrGlob);
        // build the query options and perform the request
        let baseQuery = { autoPaginate: true };
        let getListOpts = (prefix) ? Object.assign(Object.assign({}, baseQuery), { prefix }) : baseQuery;
        const result = await this.googleBucket.getFiles(getListOpts);
        let gfList = result[0] || [];
        // if glob, filter the data further
        let files = (!glob) ? gfList : gfList.filter(gf => micromatch.isMatch(gf.name, glob));
        return files;
    }
}
